# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_df_connection.spark_hdfs.connection import SparkHDFS
from onetl.connection.file_df_connection.spark_hdfs.slots import SparkHDFSSlots
