# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_connection.hdfs.connection import HDFS
from onetl.connection.file_connection.hdfs.slots import HDFSSlots
