# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.teradata.connection import Teradata, TeradataExtra
from onetl.connection.db_connection.teradata.dialect import TeradataDialect
